# -*- coding: utf-8 -*-
require 'sidfmdlm_api'
require 'sidfm_controller_support'
require 'sidfm_licenses_controller_support'

class SidfmProbesController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception
  helper :sidfm

  private

  include SidfmControllerSupport, SidfmLicensesControllerSupport
  before_action :update_session_updated_at
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_roles
  before_action :get_vm_version

  # before_action :clear_import_session, only: [ :create, :modify, :delete, export, :read_import_file ]
  before_action :clear_export_session, only: [ :index ]
  before_action -> { get_summary_table_lines('probes') }

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:host_admin) then
      redirect_to home_path, :notice => af_log_user('アクセス権限がありません。') and return
      
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = @role_admin ? nil : User.current.categories.order(:id).pluck(:id)
    @user_host_ids = @role_admin ? nil : SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.pluck(:sidfmd_host_id)
    # @apikey設定
    @apikeys = SidfmdProbeApikey.all
    unless @role_admin then
      @apikeys = @apikeys.
        where(sidfmd_category_id: @user_category_ids)
    end
    @apikeys = @apikeys.
      order("name COLLATE \"C\"")
    if @apikeys.count == 0 then
      # probe_apikeyが無い場合
      redirect_to home_path, :notice => af_log_user('探索APIキーがありません。') and return
    end
  end

  # ホスト名として正しいものか確認。
  def check_hostname?(hostname)
    return false unless hostname
    if hostname.size == 0 then
      return false
    elsif /\A[0-9a-zA-Z._-]+\z/ !~ hostname then
      return false
    elsif hostname.bytesize > SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH then
      return false
    end
    return true
  end

  def get_user_categories
    categories = @role_admin ? SidfmdCategory : SidfmdCategory.where(:id => @user_category_ids)
    categories = categories.where(:is_system => false)
  end

  def get_all_content
    # 万が一@apikeyが無ければ、空応答
    return SidfmdProbeSshIp.where(:ipaddress => nil) unless @apikey

    # 現時点の時間
    time_current = Time.current

    ### アクセス制限

    # アクセス範囲の制限
    filter_probe = -> (q) {
      q.
      where(sidfmd_probes: { sidfmd_probe_apikey_id: @apikey.id }).
      where(sidfmd_probes: { flg_enable: true }).
      where(sidfmd_probes: { kind: 'ssh' })
    }

    # ipアドレス毎に値を決定
    filter_probe_ssh = -> (q) {
      q.
      where("sidfmd_probe_ssh_ips.ipaddress=pre_ips.ipaddress")
    }

    ### テーブル指定
    probe_ssh = -> {
      filter_probe.
      call(SidfmdProbeSsh.
           joins(:sidfmd_probe))
    }

    probe_ssh_ip = -> (filter_ssh = true) {
      q = filter_probe.call(SidfmdProbeSshIp.
                            joins(:sidfmd_probe_ssh).
                            joins(sidfmd_probe_ssh: :sidfmd_probe))
      if filter_ssh then
        q = filter_probe_ssh.call(q)
      end
      q
    }

    probe_ssh_ip_nmap = -> {
      filter_probe_ssh.
      call(filter_probe.
           call(SidfmdProbeSshIpNmap.
                joins(:sidfmd_probe_ssh_ip).
                joins(sidfmd_probe_ssh_ip: :sidfmd_probe_ssh).
                joins(sidfmd_probe_ssh_ip: { sidfmd_probe_ssh: :sidfmd_probe })))
    }

    probe_ssh_ip_pkg = -> {
      filter_probe_ssh.
      call(filter_probe.
           call(SidfmdProbeSshIpPkg.
                joins(:sidfmd_probe_ssh_ip).
                joins(sidfmd_probe_ssh_ip: :sidfmd_probe_ssh).
                joins(sidfmd_probe_ssh_ip: { sidfmd_probe_ssh: :sidfmd_probe })))
    }


    # 最新のdatekey
    @last_datekey =
      probe_ssh.call.
      select(:datekey).
      order(:datekey => :desc).
      limit(1).
      first&.datekey

    # 対象ipアドレス一覧
    query_ips =
      filter_probe.call(probe_ssh_ip.call(false)).
      select(:ipaddress).
      group(:ipaddress).
      to_sql

    # 現在pingが通っているかどうか
    query_is_last_ping =
      probe_ssh_ip.call.
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      select(:is_ping).
      to_sql

    # 現在nmapが通っているかどうか
    query_is_last_nmap =
      probe_ssh_ip_nmap.call.
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      select("COUNT(*)>0").
      to_sql
    
    # 現在sshの情報取得が成功しているかどうか
    query_is_ssh =
      probe_ssh_ip.call.
      select("COUNT(*)>0").
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      where.not(:ident => nil).
      where.not(:version => nil).
      where("COALESCE(hostname_f,hostname_s) IS NOT NULL").
      to_sql

    #  hostname
    query_hostname =
      probe_ssh_ip.call.
      select("COALESCE(hostname_f,hostname_s)").
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      limit(1).
      to_sql

    #  ident
    query_ident =
      probe_ssh_ip.call.
      select(:ident).
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      limit(1).
      to_sql

    #  version
    query_version =
      probe_ssh_ip.call.
      select(:version).
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      limit(1).
      to_sql

    #  name
    query_name =
      probe_ssh_ip.call.
      select(:name).
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      limit(1).
      to_sql

    #  pkg
    query_pkg_count =
      probe_ssh_ip_pkg.call.
      select("COUNT(*)").
      where(sidfmd_probe_sshes: { datekey: @last_datekey }).
      to_sql
    
    # ホストとの紐づけがあるかどうか
    # nilなら紐づけなし。
    # hostnameのホストが登録後別のカスタムカテゴリになっている可能性あり。
    # 出力する際は要チェック。
    query_host = SidfmdHost.
      joins(:sidfmd_categories).
      joins(:sidfmd_probe_apikey).
      select(SidfmdHost.arel_table[:id]).
      where(flg_delete: false).
      where(sidfmd_probe_apikey_id: @apikey.id).
      where("probe_key=pre_ips.ipaddress").
      limit(1).
      to_sql

    # IPアドレス毎の最新datekey
    # IPアドレス毎のデータ送信日時相当として扱う
    query_ldt =
      probe_ssh_ip.call.
      select("sidfmd_probe_sshes.datekey").
      order("sidfmd_probe_sshes.datekey DESC").
      limit(1).
      to_sql

    # 以前はpingが通っているかどうか
    query_is_elapsed_ping =
      probe_ssh_ip.call.
      select(:is_ping).
      where("sidfmd_probe_sshes.datekey < ?", @last_datekey).
      where("sidfmd_probes.created_at < ?", time_current - Sidfm.probe_elapsed_days.days).
      order("sidfmd_probe_sshes.datekey DESC").
      limit(1).
      to_sql

    # 以前はnmapが通っているかどうか
    query_is_elapsed_nmap =
      probe_ssh_ip_nmap.call.
      where("sidfmd_probe_sshes.datekey < ?", @last_datekey).
      where("sidfmd_probes.created_at < ?", time_current - Sidfm.probe_elapsed_days.days).
      select("COUNT(*)>0").
      to_sql

    # 以前はsshの情報取得が成功しているかどうか
    query_is_elapsed_ssh =
      probe_ssh_ip.call.
      where("sidfmd_probe_sshes.datekey < ?", @last_datekey).
      where("sidfmd_probes.created_at < ?", time_current - Sidfm.probe_elapsed_days.days).
      select("COUNT(*)>0").
      where.not(:ident => nil).
      where.not(:version => nil).
      where("COALESCE(hostname_f,hostname_s) IS NOT NULL").
      to_sql
    
    # 検出されたipアドレス一覧
    query_pre_ips = SidfmdProbeSshIp.
      from("(#{query_ips}) AS pre_ips").
      select("pre_ips.ipaddress AS ipaddress").
      select("COALESCE((#{query_is_last_ping}),false) AS is_last_ping").
      select("COALESCE((#{query_is_last_nmap}),false) AS is_last_nmap").
      select("COALESCE((#{query_is_ssh}),false) AS is_last_ssh").
      select("(#{query_hostname}) AS hostname").
      select("(#{query_ident}) AS ident").
      select("(#{query_version}) AS version").
      select("(#{query_name}) AS name").
      select("(#{query_pkg_count}) AS pkg_count").
      select("(#{query_host}) AS sidfmd_host_id").
      select("(#{query_ldt}) AS ldatekey").
      select("COALESCE((#{query_is_elapsed_ping}),false) AS is_elapsed_ping").
      select("COALESCE((#{query_is_elapsed_nmap}),false) AS is_elapsed_nmap").
      select("COALESCE((#{query_is_elapsed_ssh}),false) AS is_elapsed_ssh").
      to_sql
    SidfmdProbeSshIp.from("(#{query_pre_ips}) AS ips").select("*")
  end

  def create_summary_probe_html(probe_ssh_ip, **opts)
    pkgs = []
    if @apikey and probe_ssh_ip.ipaddress and @last_datekey then
      pkgs = SidfmdProbeSshIpPkg.
        joins(:sidfmd_probe_ssh_ip).
        joins(sidfmd_probe_ssh_ip: :sidfmd_probe_ssh).
        joins(sidfmd_probe_ssh_ip: { sidfmd_probe_ssh: :sidfmd_probe }).
        where(sidfmd_probes: { flg_enable: true }).
        where(sidfmd_probes: { sidfmd_probe_apikey_id: @apikey.id }).
        where(sidfmd_probe_sshes: { datekey: @last_datekey }).
        where(sidfmd_probe_ssh_ips: { ipaddress: probe_ssh_ip.ipaddress }).
        select(:pkg).
        order(:pkg).
        pluck(:pkg)
    end
    sidfmd_host_hostname = nil
    if probe_ssh_ip.sidfmd_host_id then
      q = SidfmdHost.
        where(flg_delete: false).
        where(id: probe_ssh_ip.sidfmd_host_id)
      unless @role_admin then
        q = q.
          joins(:sidfmd_categories).
          where(sidfmd_categories: { id: @user_category_ids })
      end
      if q.count > 0 then
        sidfmd_host_hostname = q.first.hostname
      end
    end
    data = {
      :chkbox => view_context.create_summary_probe_chkbox(probe_ssh_ip.ipaddress, opts[:checked]),
      :ipaddress => view_context.create_summary_probe_ipaddress(probe_ssh_ip.ipaddress),
      :hostname => view_context.create_summary_probe_hostname(probe_ssh_ip.hostname),
      :datekey => view_context.create_summary_probe_datekey(probe_ssh_ip.ldatekey),
      :is_ssh => view_context.create_summary_probe_is_ssh(probe_ssh_ip.is_last_ssh, probe_ssh_ip.is_elapsed_ssh),
      :is_nmap => view_context.create_summary_probe_is_nmap(probe_ssh_ip.is_last_nmap, probe_ssh_ip.is_elapsed_nmap),
      :is_ping => view_context.create_summary_probe_is_ping(probe_ssh_ip.is_last_ping, probe_ssh_ip.is_elapsed_ping),
      :os => view_context.create_summary_probe_os(probe_ssh_ip.ident, probe_ssh_ip.name, probe_ssh_ip.version),
      :pkgs => view_context.create_summary_probe_pkgs(pkgs),
      :edit => view_context.create_summary_probe_edit(@apikey&.id, probe_ssh_ip.is_last_ssh, probe_ssh_ip.sidfmd_host_id, sidfmd_host_hostname, probe_ssh_ip.ipaddress, probe_ssh_ip.hostname, @last_datekey),
    }
  end

  def parse_params
    f_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :filters, :checked, :all_checked, :probe_apikey_id)
    @probes = {}
    @probes[:per_page] = (@summary_table_lines || f_params[:per_page] || 10).to_i
    @probes[:offset] = (f_params[:offset] || 0).to_i
    @probes[:order] = f_params[:order] || 'ipaddress_asc'
    @probes[:filters] = JSON.parse((f_params[:filters] || '{}'), symbolize_names: true)
    @probes[:checked] = JSON.parse(f_params[:checked] || '[]')
    @probes[:all_checked] = f_params[:all_checked] || false
    # apikey
    @apikey = nil
    # pages_probe_apikey_idには変更後正しい値が出力されません。
    # @probes[:probe_apikey_id] = (f_params[:probe_apikey_id] || 0).to_i
    @probes[:probe_apikey_id] = (@probes[:filters][:probe_apikey_id] || 0).to_i
    if @probes[:probe_apikey_id] and @probes[:probe_apikey_id] > 0 then
      q = SidfmdProbeApikey.
        where(:id => @probes[:probe_apikey_id])
      unless @role_admin then
        q = q.
          where(sidfmd_category_id: @user_category_ids)
      end
      if q.count > 0 then
        @apikey = q.first
      else
        @apikey = @apikeys.first
      end
    else
      @apikey = @apikeys.first
    end
    @probes[:probe_apikey_id] = @apikey.id.to_s
    @default_probe_apikey_id = @apikeys.first.id
  end

  FILTER_KEYS = {}

  def init_filter_keys
    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:is_ssh] = ''
      FILTER_KEYS[:relate_host] = ''
      FILTER_KEYS[:probe_apikey_id] = @default_probe_apikey_id.to_s
    end
  end

  def check_filters
    init_filter_keys
    FILTER_KEYS.each do |key, val|
      unless @probes[:filters].has_key?(key) then
        @probes[:filters][key] = val
      end
    end
    @default_filters = FILTER_KEYS
  end

  def sort_content
    order_data = @probes[:order].split('_')
    if order_data.size > 2 then
      col = "#{order_data[0]}_#{order_data[1]}"
    else
      col = order_data.first
    end
    order = SidfmControllerSupport.check_order(order_data.last)

    case col
    when 'ipaddress' then
      @probes[:all_content] = @probes[:all_content].
        order("ipaddress::inet #{order.upcase}")
    when 'hostname' then
      @probes[:all_content] = @probes[:all_content].
        order("hostname #{order.upcase}")
    when 'datekey' then
      @probes[:all_content] = @probes[:all_content].
        order("ldatekey #{order.upcase}")
    when 'is_ssh' then
      @probes[:all_content] = @probes[:all_content].
        order("is_last_ssh #{order.upcase}").
        order("is_elapsed_ssh #{order.upcase}")
    when 'is_nmap' then
      @probes[:all_content] = @probes[:all_content].
        order("is_last_nmap #{order.upcase}").
        order("is_elapsed_nmap #{order.upcase}")
    when 'is_ping' then
      @probes[:all_content] = @probes[:all_content].
        order("is_last_ping #{order.upcase}").
        order("is_elapsed_ping #{order.upcase}")
    when 'os' then
      @probes[:all_content] = @probes[:all_content].
        order("name #{order.upcase}").
        order("version #{order.upcase}")
    when 'pkgs' then
      @probes[:all_content] = @probes[:all_content].
        order("pkg_count #{order.upcase}")
    end

  end

  def filter_content
    if @probes[:filters][:probe_apikey_id].size > 0 then
      new_apikey_id = @probes[:filters][:probe_apikey_id].to_i
      if @apikey.id != new_apikey_id and @apikeys.pluck(:id).include?(new_apikey_id) then
        @apikey = @apikeys.find {|x| x.id == new_apikey_id }
        @probes[:all_content] = get_all_content
      else
        @probes[:filters][:probe_apikey_id] = @apikey.id.to_s
      end
    end
    if @probes[:filters][:is_ssh].size > 0 then
      case @probes[:filters][:is_ssh]
      when 'success' then
        @probes[:all_content] = @probes[:all_content].
          where(ips: { is_last_ssh: true })
      when 'failure' then
        @probes[:all_content] = @probes[:all_content].
          where(ips: { is_last_ssh: false })
      when 'success<-failure' then
        @probes[:all_content] = @probes[:all_content].
          where(ips: { is_last_ssh: true }).
          where(ips: { is_elapsed_ssh: false })
      when 'failure<-success' then
        @probes[:all_content] = @probes[:all_content].
          where(ips: { is_last_ssh: false }).
          where(ips: { is_elapsed_ssh: true })
      else
        @probes[:filters][:is_ssh] = ''
      end
    end
    if @probes[:filters][:relate_host].size > 0 then
      case @probes[:filters][:relate_host]
      when 'enable' then
        @probes[:all_content] = @probes[:all_content].
          where.not(ips: { sidfmd_host_id: nil })
      when 'disable' then
        @probes[:all_content] = @probes[:all_content].
          where(ips: { sidfmd_host_id: nil })
      else
        @probes[:filters][:relate_host] = ''
      end
    end
  end

  public

  def index
    parse_params
    check_filters
    @page_name = '探索済ホスト設定'
    @user_categories = get_user_categories
  end

  def api_get_display_summary_content
    GC.start
    f_params = params.fetch(:pages, {}).permit(:per_page)
    parse_params
    @probes[:per_page] = f_params[:per_page].to_i
    @probes[:all_content] = get_all_content
    check_filters
    filter_content
    sort_content
    @probes[:content_filtered] = @probes[:all_content]
    ipaddresses = @probes[:content_filtered].map(&:ipaddress)
    #create_cat_group_categories(ids)
    @probes[:content] = @probes[:content_filtered].limit(@probes[:per_page]).offset(@probes[:offset])
    user_id = User.current.id
    ret = {
      :rows => @probes[:content].map{|probe_ssh_ip|
        create_summary_probe_html(
                                  probe_ssh_ip,
                                  {
                                    :user_id => user_id,
                                    :checked => @probes[:checked],
                                    :last_datekey => @last_datekey,
                                  }
                                  )
      },
      :total => ipaddresses.size,
      :all_ids => ipaddresses,
      :sidfm_roles => @sidfm_roles,
      :user_cats => @user_category_ids,
    }
    render :layout => false, :json => ret
  end

  # 紐づけの削除
  def api_delete
    f_params = params.permit(:probe_apikey_id, :ipaddress, :sidfmd_host_id)

    probe_apikey_id = f_params[:probe_apikey_id].to_i
    ipaddress = f_params[:ipaddress]
    sidfmd_host_id = f_params[:sidfmd_host_id].to_i

    # ipaddress check
    unless /\A(\d+)\.(\d+).(\d+)\.(\d+)\z/ =~ ipaddress then
      render :layout => false, :json => { :err_message => 'invalid ipaddress' } and return
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    # sidfmd_host_id check
    q_check_hostid = SidfmdHost.
      where(id: sidfmd_host_id).
      where(sidfmd_probe_apikey_id: probe_apikey.id).
      where(probe_key: ipaddress)
    unless @role_admin then
      q_check_hostid = q_check_hostid.where(:id => @user_host_ids)
    end
    if q_check_hostid.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid sidfmd_host_id' } and return
    end
    sidfmd_host = q_check_hostid.first

    SidfmdHost.transaction do
      sidfmd_host.lock!
      sidfmd_host.update_attribute :sidfmd_probe_apikey_id, nil
      sidfmd_host.update_attribute :probe_key, nil
    end

    render :layout => false, :json => { :err_message => '' }
  end

  # 一括紐づけ削除時の一覧表を返す。
  def api_get_ip_hostname
    f_params = params.permit(:ips, :probe_apikey_id)
    probe_apikey_id = f_params[:probe_apikey_id]

    # ips check
    ips = nil
    begin
      ips = JSON.parse(f_params[:ips] || '[]')
    rescue JSON::ParserError
      ips = []
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    ret = []
    ips.each do |ip|
      sidfmd_host = SidfmdHost.
        where(flg_delete: false).
        where(sidfmd_probe_apikey_id: probe_apikey.id).
        where(probe_key: ip)
      unless @role_admin then
        sidfmd_host = sidfmd_host.where(id: @user_host_ids)
      end
      if sidfmd_host.count > 0 then
        ret << { ipaddress: ip, hostname: sidfmd_host.first.hostname }
      end
    end

    render :layout => false, :json => ret
  end

  # 一括ホスト登録時の一覧表を返す。
  def api_get_ip_hostname_regist
    f_params = params.permit(:ips, :probe_apikey_id)
    probe_apikey_id = f_params[:probe_apikey_id]

    # ips check
    ips = nil
    begin
      ips = JSON.parse(f_params[:ips] || '[]')
    rescue JSON::ParserError
      ips = []
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    ret = []
    ips.each do |ip|
      q_check_duphost = SidfmdHost.
        where(flg_delete: false).
        where(sidfmd_probe_apikey_id: probe_apikey.id).
        where(probe_key: ip)
      if q_check_duphost.count == 0 then
        datekey = SidfmdProbeSsh.
          joins(:sidfmd_probe).
          joins(:sidfmd_probe_ssh_ips).
          where(sidfmd_probes: { flg_enable: true }).
          where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).
          where(sidfmd_probe_ssh_ips: { ipaddress: ip }).
          order(:datekey => :desc).
          limit(1).
          first&.
          datekey
        if datekey then
          ssh_ip = SidfmdProbeSshIp.
            joins(:sidfmd_probe_ssh).
            joins(:sidfmd_probe_ssh => :sidfmd_probe).
            where(ipaddress: ip).
            where.not(ident: nil).
            where.not(version: nil).
            where("COALESCE(hostname_f,hostname_s) IS NOT NULL").
            where(sidfmd_probe_sshes: { datekey: datekey }).
            where(sidfmd_probes: { flg_enable: true }).
            where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).
            first
          if ssh_ip then
            hostname = (ssh_ip.hostname_f || ssh_ip.hostname_s)
            ret << { ipaddress: ip, hostname: hostname }
          else
            logger.info(" skip: missing ssh_ip #{ip}")
          end
        else
          logger.info(" skip: missing datekey #{ip}")
        end
      else
        logger.info(" skip: duphost #{ip}")
      end
    end

    render :layout => false, :json => ret
  end

  # 一括紐づけ削除
  def api_bulk_delete
    f_params = params.permit(:probe_apikey_id, :ips)

    probe_apikey_id = f_params[:probe_apikey_id].to_i

    # ips check
    ips = nil
    begin
      ips = JSON.parse(f_params[:ips] || '[]')
    rescue JSON::ParserError => e
      log_exception_(e)
      ips = []
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    ret = []
    ips.each do |ip|
      sidfmd_host = SidfmdHost.
        where(flg_delete: false).
        where(sidfmd_probe_apikey_id: probe_apikey.id).
        where(probe_key: ip)
      unless @role_admin then
        sidfmd_host = sidfmd_host.where(id: @user_host_ids)
      end
      if sidfmd_host.count > 0 then
        SidfmdHost.transaction do
          sidfmd_host_f = sidfmd_host.first
          sidfmd_host_f.lock!
          sidfmd_host_f.update_attribute :sidfmd_probe_apikey_id, nil
          sidfmd_host_f.update_attribute :probe_key, nil
        end
      end
    end
    Rails.cache.clear

    render :layout => false, :json => { :err_message => '' }
  end

  # ホスト名に重複が無いかチェック
  #   重複しているホスト名を配列で返す。
  def api_check_duplicate_hostnames
    f_params = params.permit(:hostnames);

    hostnames = nil
    begin
      hostnames = JSON.parse(f_params[:hostnames] || '[]')
    rescue JSON::ParserError => e
      log_exception(e)
    end

    ret = []
    if hostnames then
      ret = SidfmdHost.
        where(flg_delete: false).
        where(hostname: hostnames).
        pluck(:hostname)
      if ret.size == 0 then
        ret = hostnames.collect {|x| check_hostname?(x) ? nil : x }.compact
      end
    else
      ret << '--- invalid param: hostnames ---'
    end

    render :layout => false, :json => ret
  end

  # 一括ホスト登録時のホスト名を取得
  # 以下の形式で返す。
  # [ { ipaddress: 'xxx.xxx.xxx.xxx', hostname: 'xxxx' }, ... ]
  def api_get_generate_hostnames
    f_params = params.permit(:probe_apikey_id, :ips, :prefix, :suffix, :iphost)

    probe_apikey_id = f_params[:probe_apikey_id].to_i
    prefix = (f_params[:prefix] || '')
    suffix = (f_params[:suffix] || '')
    iphost = (f_params[:iphost] || 'hostname')
    unless [ 'ip', 'hostname' ].include?(iphost) then
      iphost = 'hostname'
    end

    # ips check
    ips = nil
    begin
      ips = JSON.parse(f_params[:ips] || '[]')
      unless ips.respond_to?(:[]) then
        log.error("invalid ips: #{ips.class}")
        render :layout => false, :json => { :err_message => 'invalid ips' } and return
      end
    rescue JSON::ParserError => e
      log_exception_(e)
      render :layout => false, :json => { :err_message => 'invalid ips' } and return
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    ret = []
    if 'ip' == iphost then
      ips.each do |ip|
        hostname = "#{prefix}#{ip}#{suffix}"
        logger.info("#{ip} #{hostname} #{check_hostname?(hostname)}")
        unless check_hostname?(hostname) then
          render :layout => false, :json => JSON.generate([]) and return
        end
        ret << { ipaddress: ip, hostname: hostname }
      end
    elsif 'hostname' == iphost then
      ips.each do |ip|
        datekey = SidfmdProbeSsh.
          joins(:sidfmd_probe).
          joins(:sidfmd_probe_ssh_ips).
          where(sidfmd_probes: { flg_enable: true }).
          where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).
          where(sidfmd_probe_ssh_ips: { ipaddress: ip }).
          order(:datekey => :desc).
          limit(1).
          first&.
          datekey
        if datekey then
          ssh_ip = SidfmdProbeSshIp.
            joins(:sidfmd_probe_ssh).
            joins(:sidfmd_probe_ssh => :sidfmd_probe).
            where(ipaddress: ip).
            where(sidfmd_probe_sshes: { datekey: datekey }).
            where(sidfmd_probes: { flg_enable: true }).
            where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).first
          if ssh_ip then
            hostname = (ssh_ip.hostname_f || ssh_ip.hostname_s)
            if hostname then
              gen_hostname = "#{prefix}#{hostname}#{suffix}"
              if check_hostname?(gen_hostname) then
                ret << { ipaddress: ip, hostname: gen_hostname }
              else
                render :layout => false, :json => JSON.generate([]) and return
              end
            else
              logger.error("unknown error: hostname : #{ip}")
              render :layout => false, :json => JSON.generate([]) and return
            end
          else
            logger.error("unknown error: ssh_ip : #{ip}")
            render :layout => false, :json => JSON.generate([]) and return
          end
        else
          logger.error("unknown error: datekey : #{ip}")
          render :layout => false, :json => JSON.generate([]) and return
        end
      end
    end

    render :layout => false, :json => ret
  end

  # ホスト登録
  # 一括ホスト登録
  def api_regist
    f_params = params.permit(:probe_apikey_id, :ips);

    probe_apikey_id = f_params[:probe_apikey_id].to_i

    # ips check
    ips = nil
    begin
      ips = JSON.parse(f_params[:ips] || '[]')
    rescue JSON::ParserError => e
      log_exception_(e)
      render :layout => false, :json => { :err_message => 'invalid ips' } and return
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    # ホスト登録
    system_category_id = 1
    refdate = Date.today.to_s
    err_ips = []
    ips.each do |ip|
      ip_ipaddress = ip['ipaddress']
      ip_hostname = ip['hostname']
      if ip_ipaddress and ip_hostname and check_hostname?(ip_hostname) then
        q_check_duphost = SidfmdHost.
          where(flg_delete: false).
          where(sidfmd_probe_apikey_id: probe_apikey.id).
          where(probe_key: ip_ipaddress)
        if q_check_duphost.count > 0 then
          logger.info(" skip: error create_host #{ip.inspect}")
          err_ips << ip
          next
        end
        res = create_host(hostname: ip_hostname,
                          display_name: nil,
                          system_category_id: system_category_id,
                          custom_category_ids: [ probe_apikey.sidfmd_category_id ],
                          refdate: refdate,
                          memo: '',
                          host_apps: [],
                          host_lapps: [],
                          role_admin: @role_admin,
                          user_category_ids: @user_category_ids,
                          sidfm_roles: @sidfm_roles)
        if res[:success] then
          SidfmdHost.transaction do
            sidfmd_host = SidfmdHost.find_by_hostname(ip_hostname)
            if sidfmd_host then
              sidfmd_host.lock!
              datekey = SidfmdProbeSsh.
                joins(:sidfmd_probe).
                joins(:sidfmd_probe_ssh_ips).
                where(sidfmd_probes: { flg_enable: true }).
                where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).
                where(sidfmd_probe_ssh_ips: { ipaddress: ip_ipaddress }).
                order(:datekey => :desc).
                limit(1).
                first&.
                datekey
              if datekey then
                opts = { :os => {}, :packages => [], :applications => [] }
                ssh_ip = SidfmdProbeSshIp.
                  joins(:sidfmd_probe_ssh).
                  joins(:sidfmd_probe_ssh => :sidfmd_probe).
                  where(ipaddress: ip_ipaddress).
                  where(sidfmd_probe_sshes: { datekey: datekey }).
                  where(sidfmd_probes: { flg_enable: true }).
                  where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).first
                if ssh_ip then
                  opts[:os][:id] = ssh_ip.ident
                  opts[:os][:version] = ssh_ip.version
                  opts[:os][:os] = ssh_ip.name
                  opts[:packages] = ssh_ip.sidfmd_probe_ssh_ip_pkgs.order(:pkg).pluck(:pkg)
                  queue = sidfmd_host.create_queue(opts)
                end
              end
              sidfmd_host.update_attribute :is_agent, true
              sidfmd_host.update_attribute :sidfmd_probe_apikey_id, probe_apikey.id
              sidfmd_host.update_attribute :probe_key, ip_ipaddress
            end
          end
        else
          logger.info(" skip: error create_host #{ip.inspect}")
          err_ips << ip
        end
                          
      else
        logger.info(" skip: invalid ips #{ip.inspect}")
        err_ips << ip
      end
    end
    Rails.cache.clear

    if err_ips.size > 0 then
      render :layout => false, :json => { :err_message => "error create_host #{err_ips.size} hosts" } and return
    end

    render :layout => false, :json => { :err_message => '' }
  end

  # ホスト検索用ホスト取得api
  def api_search_hostnames
    f_params = params.permit(:hostname)
    hostname = (f_params[:hostname] || '')

    escaped_hostname = "%#{ActiveRecord::Base.send(:sanitize_sql_like, hostname)}%"
    q = SidfmdHost.
      where(flg_delete: false).
      where(sidfmd_probe_apikey_id: nil).
      select(:id).
      select(:hostname).
      select(:ext_name).
      where("hostname ILIKE ? OR ext_name ILIKE ?", escaped_hostname, escaped_hostname).
      order(:hostname)
    unless @role_admin then
      q = q.where(id: @user_host_ids)
    end
    render :layout => false, :json => q.map {|x| { id: x.id, hostname: x.hostname, displayname: x.ext_name }}
  end
  
  # ホスト紐づけ実行
  def api_edit
    f_params = params.permit(:probe_apikey_id, :ips)
    # ips = [ { ipaddress, hostid } ]

    probe_apikey_id = f_params[:probe_apikey_id].to_i

    # ips check
    ips = nil
    begin
      ips = JSON.parse(f_params[:ips] || '[]')
    rescue JSON::ParserError => e
      log_exception_(e)
      render :layout => false, :json => { :err_message => 'invalid ips' } and return
    end

    # probe_apikey_id check
    q_check_apikey = SidfmdProbeApikey.where(id: probe_apikey_id.to_i)
    unless @role_admin then
      q_check_apikey = q_check_apikey.where(sidfmd_category_id: @user_category_ids)
    end
    if q_check_apikey.count == 0 then
      render :layout => false, :json => { :err_message => 'invalid probe_apikey_id' } and return
    end
    probe_apikey = q_check_apikey.first

    # ホスト紐づけ
    err_ips = []
    ips.each do |ip|
      ip_ipaddress = ip['ipaddress']
      ip_hostid = ip['hostid'].to_i
      q_check_hostid = SidfmdHost.
        where(flg_delete: false).
        where(sidfmd_probe_apikey_id: nil).
        where(id: ip_hostid)
      unless @role_admin then
        q_check_hostid = q_check_hostid.
          where(id: @user_host_ids)
      end
      if q_check_hostid.count > 0 then
        SidfmdHost.transaction do
          sidfmd_host = q_check_hostid.first
          sidfmd_host.lock!
          datekey = SidfmdProbeSsh.
            joins(:sidfmd_probe).
            joins(:sidfmd_probe_ssh_ips).
            where(sidfmd_probes: { flg_enable: true }).
            where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).
            where(sidfmd_probe_ssh_ips: { ipaddress: ip_ipaddress }).
            order(:datekey => :desc).
            limit(1).
            first&.
            datekey
          if datekey then
            opts = { :os => {}, :packages => [], :applications => [] }
            ssh_ip = SidfmdProbeSshIp.
              joins(:sidfmd_probe_ssh).
              joins(:sidfmd_probe_ssh => :sidfmd_probe).
              where(ipaddress: ip_ipaddress).
              where(sidfmd_probe_sshes: { datekey: datekey }).
              where(sidfmd_probes: { flg_enable: true }).
              where(sidfmd_probes: { sidfmd_probe_apikey_id: probe_apikey.id }).first
            if ssh_ip then
              opts[:os][:id] = ssh_ip.ident
              opts[:os][:version] = ssh_ip.version
              opts[:os][:os] = ssh_ip.name
              opts[:packages] = ssh_ip.sidfmd_probe_ssh_ip_pkgs.order(:pkg).pluck(:pkg)
              queue = sidfmd_host.create_queue(opts)
            end
            sidfmd_host.update_attribute :probe_key, ip_ipaddress
            sidfmd_host.update_attribute :sidfmd_probe_apikey_id, probe_apikey.id
            sidfmd_host.update_attribute :is_agent, true
          end
        end
      else
        logger.info(" skip: not found hostid: #{ip.inspect}")
        err_ips << ip
      end
    end
    Rails.cache.clear

    if err_ips.size > 0 then
      render :layout => false, :json => { :err_message => "error relate_host #{err_ips.size} hosts" } and return
    end

    render :layout => false, :json => { :err_message => '' }
  end

end
