#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

module SidfmLicensesControllerSupport
  include SidfmVmApiControllerSupport
  # 既存のメソッドとメソッド名が被らないように注意

  # 登録・更新時のvalidation
  # update_host_idは更新の時にのみ、対象のホストidを指定する
  def validate_host_params(opts:, update_host_id: nil, role_admin:, user_category_ids:)
    err_messages = []
    if !update_host_id || opts.has_key?(:hostname) then
      hostname = opts[:hostname]
      hostname = '' unless hostname
      if hostname.size == 0 then
        err_messages << 'ホスト名が入力されていません。'
      elsif /\A[0-9a-zA-Z._-]+\z/ !~ hostname then
        err_messages << 'ホスト名が不正です。'
      elsif hostname.bytesize > SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH then
        err_messages << SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH.to_s + 'バイトを超えるホスト名は登録できません。'
      end
    end
    if !update_host_id || opts.has_key?(:display_name) then
      display_name = opts[:display_name]
      if display_name then
        if display_name.size > SidfmdHost::MAX_DISPLAY_NAME_LENGTH then
          err_messages << '100文字を超える表示名は登録できません。'
        elsif display_name =~ /<|>|"|'/ then
          err_messages << '表示名に使用できない文字が含まれています。'
        elsif display_name.size > 0 && display_name.strip.size === 0 then
          err_messages << '空白文字のみの表示名は登録できません。'
        elsif display_name =~ /&&&&&/ then
          err_messages << '表示名に「&&&&&」の文字列(5つ連なるアンド)は使用できません。'
        end
      end
    end

    if !update_host_id || opts.has_key?(:system_category_id) then
      system_category_id = opts[:system_category_id].to_i
      if !SidfmdCategory.where(:id => system_category_id).where(:is_system => true).exists? then
        err_messages << 'システムカテゴリが不正です。'
      end
    end

    if !update_host_id || opts.has_key?(:custom_category_ids) then
      custom_category_ids = opts[:custom_category_ids]
      if !role_admin && custom_category_ids.size == 0 then
        err_messages << 'カスタムカテゴリが設定されていません。'
      elsif !role_admin then
        diff = custom_category_ids & user_category_ids
        if diff.size == 0 then
          err_messages << '無効なカスタムカテゴリです。'
        end
      end
      custom_category_ids = [] unless custom_category_ids
      hostlim_over_categories = SidfmdCategoriesHost.hostlim_over_categories(
        { (update_host_id ? update_host_id.to_i : 0) => custom_category_ids }
      )
      if hostlim_over_categories.any? then
        err_messages << "次のカテゴリは、ホスト登録数が上限を超えるため、登録できません。 #{hostlim_over_categories.map{|cat| cat.name}.join(", ")}"
      end
    end

    if !update_host_id || opts.has_key?(:refdate) then
      refdate = opts[:refdate]
      if refdate && refdate.size > 0 then
        begin
          refdate = Date.parse(refdate)
        rescue ArgumentError => e
          log_exception(e)
          err_messages << '基準日に不正な日付が指定されています。'
        end
      else
        if !update_host_id then #更新のみ基準日を削除可(UIと同じ)
          err_messages << '基準日が指定されていません。'
        end
      end
    end

    if !update_host_id || opts.has_key?(:memo) then
      memo = opts[:memo]
      memo = '' unless memo
      if memo.length > SidfmdHost::MEMO_MAX_LEN then
        err_messages << 'メモに入力可能な最大文字数を超えています'
      end
    end

    host_apps = opts[:host_apps]
    host_lapps = opts[:host_lapps]
    if host_apps && host_apps.any? then
      unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids
      host_apps.each do |app|
        if app.version_id == 0 and unregistrable_product_ids.include?(app.product_id) then
          err_messages << 'プロダクトに、「全て」を登録することができないプロダクトが含まれています。'
        end
      end
    end
    if host_apps || host_lapps then
      host_apps = [] unless host_apps
      host_lapps = [] unless host_lapps
      if (host_apps.size + host_lapps.size) > SidfmdHost::MAX_REGISTER_PRODUCTS then
        err_messages << "登録可能なプロダクトは、SIDfmプロダクトとローカルプロダクト合わせて最大#{SidfmdHost::MAX_REGISTER_PRODUCTS}個です。"
      end
    end

    packages = opts[:packages]
    package_os = opts[:package_os]
    if packages && !package_os then
      err_messages << "パッケージのOSが指定されていません。"
    elsif !packages && package_os then
      err_messages << "パッケージが指定されていません。"
    elsif packages && package_os then
      pkg_validation = validate_packages(pkgs: packages, osname: package_os)
      unless pkg_validation[:success] then
        err_messages << pkg_validation[:message]
      end
    end
    apps = []
    if opts[:enable_apps] then
      apps.push(*opts[:enable_apps])
    end
    if opts[:disable_apps] then
      apps.push(*opts[:disable_apps])
    end
    if apps.size > 0 then
      app_validation = validate_detected_apps(apps)
      unless app_validation[:success] then
        err_messages << app_validation[:message]
      end
    end
    { :success => err_messages.size == 0, :messages => err_messages }
  end

  def validate_detected_apps(apps)
    invalid_apps = apps.select{|app| !app.is_a?(String) || app.split(':').size != 3 }
    if invalid_apps.size != 0 then
      return create_process_result "指定したアプリに誤りがあります。誤りのあるアプリ: #{invalid_apps.join(', ')}"
    end
    create_process_result
  end

  def validate_packages(pkgs:, osname:)
    pkg_os_list = SidfmdQueuePkg::AMAZON_LINUX_LIST + SidfmdQueuePkg::CENTOS_LIST +
                  SidfmdQueuePkg::ALMA_LINUX_LIST +
                  SidfmdQueuePkg::DEBIAN_LIST + SidfmdQueuePkg::RHEL_LIST +
                  SidfmdQueuePkg::UBUNTU_LIST

    if !pkg_os_list.include?(osname) then
      logger.error("invalid osname: #{osname.inspect}")
      return create_process_result '指定したパッケージのOSは登録できません。'
    end

    #validation
    #jsの方で同じチェックをしているため、UIからの操作は通常問題ない
    err_pkgs = []
    regexp = if osname.match(/(Red Hat|AlmaLinux).+(8|9)/i) then
               Regexp.new(/\A[a-zA-Z0-9\._\-:\+~\(\)]+-[a-zA-Z0-9\._\-:\+~\(\)]+-[a-zA-Z0-9\._\-:\+~\(\)]+\.[a-zA-Z0-9\._\-:\+~\(\)]+\.rpm\s+[a-zA-Z0-9\._\-:\+~\(\)]+\z/)
             elsif osname.match(/Amazon/i) or osname.match(/CentOS/i) or osname.match(/(Red Hat|AlmaLinux)/i) then
               Regexp.new(/\A[a-zA-Z0-9\._\-:\+~\(\)]+-[a-zA-Z0-9\._\-:\+~\(\)]+-[a-zA-Z0-9\._\-:\+~\(\)]+\.[a-zA-Z0-9\._\-:\+~\(\)]+\.rpm\z/)
             elsif osname.match(/Ubuntu/i) or osname.match(/Debian/i) then
               Regexp.new(/\A[a-zA-Z0-9\._\-:\+~\(\)]+_[a-zA-Z0-9\._\-:\+~\(\)]+\z/)
             end
    unless regexp then
      logger.error("invalid osname: #{osname.inspect}")
      return create_process_result '指定したパッケージのOSは登録できません。'
    end

    pkgs.each do |pkg|
      if !pkg.match(regexp) then
        err_pkgs << pkg
      end
    end

    if err_pkgs.any? then
      logger.error("invalid pkgs: #{err_pkgs.inspect}")
      return create_process_result "パッケージに誤りがあります。誤りのあるパッケージ: #{err_pkgs.join(', ')}"
    end

    create_process_result
  end

  # ホストの新規登録
  def create_host(
    hostname:,
    display_name:,
    system_category_id:,
    custom_category_ids:,
    refdate:,
    memo:,
    host_apps:,
    host_lapps:,
    role_admin:,
    user_category_ids:,
    sidfm_roles:
  )
    unless sidfm_roles.include?(:host_admin) then
      return create_process_result 'forbidden access.'
    end
    validation = validate_host_params(
      opts: {
        hostname: hostname,
        display_name: display_name,
        system_category_id: system_category_id,
        custom_category_ids: custom_category_ids,
        refdate: refdate,
        memo: memo,
        host_apps: host_apps,
        host_lapps: host_lapps,
      },
      role_admin: role_admin,
      user_category_ids: user_category_ids
    )
    if !validation[:success] then
      return create_process_result validation[:messages].join
    end

    if !role_admin then
      diff = custom_category_ids & user_category_ids
      custom_category_ids = diff
    end
    
    # 既に登録されている場合はreject
    # 削除済ならそのまま通り（有効になる）。
    old_host = SidfmdHost.where(:hostname => hostname).where(:flg_delete => false)
    if old_host and old_host.count > 0 then
      return create_process_result '同名のホストが登録されています。'
    end

    _hosts = SidfmdHost.get_hosts_by_lm(:limit => 0)

    # ライセンスチェック
    # なお、ライセンスサーバ側でもチェックが行われるため、このチェックを除外しても
    # ライセンス数を超えてホストの登録を行うことはできない。
    if _hosts and _hosts[:license_number] <= _hosts[:used_license_number] then
      return create_process_result 'ホスト数がライセンスの上限に達しています。'
    end

    # ライセンスサーバとVMでホストの登録を行う。
    temp = SidfmdHost.register_host(hostname, nil, false)
    if temp.nil? then
      return create_process_result 'ホスト登録に失敗しました。'
    end

    host_id = nil
    begin
      SidfmdHost.transaction do
        host = SidfmdHost.lock.find_by(:hostname => hostname)
        if host.nil? then
         return create_process_result 'ホスト登録に失敗しました。'
        end
        host_id = host.id
        #表示名登録
        display_name = nil if display_name && display_name.size == 0
        host.update_attribute :ext_name, display_name

        # カテゴリ登録
        SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,
                                    :sidfmd_category_id => system_category_id)
        res = host.modify_categories(system_category_id, custom_category_ids)
        unless res then
          raise StandardError
        end

        # 基準日登録
        host.update_attribute :refdate, refdate

        # メモ登録
        host.update_attribute :memo, memo

        # アプリ登録
        host.sidfmd_host_apps = []
        host.sidfmd_host_lapps = []
        host.save

        if (host_apps.size + host_lapps.size) <= SidfmdHost::MAX_REGISTER_PRODUCTS then
          host_apps.each{|app| app.sidfmd_host_id = host.id }
          host_lapps.each{|lapp| lapp.sidfmd_host_id = host.id }
          SidfmdHostApp.import(host_apps) if host_apps.any?
          SidfmdHostLapp.import(host_lapps) if host_lapps.any?
        end
      end

    rescue Exception => e
      log_exception(e)
      SidfmdHostView.sync_data([ host_id ]) if host_id #Viewに反映
      return create_process_result 'ホストの表示名、カテゴリ、基準日、メモ、プロダクト、ローカルプロダクトの登録に失敗しました。'
    end
    SidfmdHostView.sync_data([ host_id ]) #Viewに反映
    Rails.cache.clear
    return create_process_result nil, { :id => host_id }
  end

  # ホストの更新
  # 引数のハッシュにあるものだけ更新する。
  # 引数はのハッシュに指定するキーは下記の通り
  #  id:, #ホストのID
  #  hostname:,
  #  display_name:,
  #  system_category_id:,
  #  custom_category_ids:,
  #  refdate:,
  #  memo:,
  #  host_apps:,
  #  host_lapps:,
  #  package_os:,
  #  packages:,
  #  enable_apps:,
  #  disable_apps:,
  #  role_admin:,
  #  user_category_ids:,
  #  sidfm_roles:
  def update_host(**opts)
    if !opts.has_key?(:id) ||
       !opts.has_key?(:sidfm_roles) ||
       !opts.has_key?(:role_admin) ||
       (!opts[:role_admin] && (!opts.has_key?(:user_category_ids) || !opts.has_key?(:user_host_ids)))
    then
      logger.info(opts)
      return create_process_result 'invalid args.'
    end

    opts[:id] = opts[:id].to_i
    if !opts[:role_admin] && (!opts[:sidfm_roles].include?(:use) || !opts[:user_host_ids].include?(opts[:id])) then
      return create_process_result 'forbidden access.'
    end

    opts[:user_category_ids] = nil if opts[:role_admin]

    host = SidfmdHost.find_by_id opts[:id]
    if !host then
      return create_process_result "not found."
    end
    is_manual_products = opts.has_key?(:host_apps) || opts.has_key?(:host_lapps)
    validation = validate_host_params(
      opts: opts,
      update_host_id: opts[:id],
      role_admin: opts[:role_admin],
      user_category_ids: opts[:user_category_ids]
    )
    if !validation[:success] then
      return create_process_result validation[:messages].join
    end

    if opts.has_key?(:packages) && !opts.has_key?(:package_os) then
      return create_process_result "パッケージを更新する場合は、パッケージのOSの指定が必須です。"
    elsif !opts.has_key?(:packages) && opts.has_key?(:package_os) then
      return create_process_result "パッケージのOSを更新する場合は、パッケージの指定が必須です。"
    end

    if opts.has_key?(:custom_category_ids) then
      if !opts[:role_admin] then
        diff = opts[:custom_category_ids] & opts[:user_category_ids]
        opts[:custom_category_ids] = diff
      end
    end

    if opts.has_key?(:hostname) then
      if SidfmdHost.where.not(:id => opts[:id]).where(:hostname => opts[:hostname]).exists? then
        return create_process_result '既に登録済みのホスト名が指定されています。'
      end
    end

    create_app_data = ->(host_id, app){
      name, version, path = app.split(':')
      {
        :sidfmd_host_id => host_id,
        :name => name,
        :version => version,
        :path => path
      }
    }

    is_host_admin_role = opts[:sidfm_roles].include?(:host_admin)
    begin
      SidfmdHost.transaction do
        host.lock!
        # ホスト名更新
        # ホスト名更新はライセンスサーバと通信してエラーになる可能性があるため、一番初めに行う
        if is_host_admin_role && opts.has_key?(:hostname) && opts[:hostname].length > 0 then
          if opts[:hostname] != host.hostname then
            license_host_data = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
            if !license_host_data || !license_host_data.has_key?(:hosts) then
              return create_process_result 'ライセンスサーバとの通信に失敗しました。'
            end
            license_host = license_host_data[:hosts].find{|l_host| l_host[:sidfmd_host] && l_host[:sidfmd_host].id == host.id }
            unless license_host then
              SidfmdHost.establish_consistency
              return create_process_result 'ホスト名の更新に失敗しました。'
            end
            res = SidfmdHost.rename_host(license_host[:id], opts[:hostname])
            if res.respond_to?(:[]) then
              if res[:success] then
                host.update_attribute :hostname, opts[:hostname]
              else
                SidfmdHost.establish_consistency
                return create_process_result '登録済のホスト名、もしくは不正な文字を指定したため、ホスト名の更新に失敗しました。'
              end
            else
              return create_process_result 'ライセンスサーバとの通信に失敗しました。'
            end
          end
        end
        # 表示名更新
        if opts.has_key?(:display_name) then
          opts[:display_name] = nil if opts[:display_name] && opts[:display_name].size == 0
          host.update_attribute :ext_name, opts[:display_name]
        end
        # カテゴリ更新
        if is_host_admin_role then
          is_c_cats_update = false
          if opts.has_key?(:system_category_id) && opts.has_key?(:custom_category_ids) then
            host.modify_categories2 opts[:system_category_id], opts[:custom_category_ids], opts[:user_category_ids]
            is_c_cats_update = true
          elsif opts.has_key?(:system_category_id) then
            host.modify_categories2(
              opts[:system_category_id],
              host.sidfmd_categories.where(:is_system => false).pluck(:id),
              opts[:user_category_ids]
            )
          elsif opts.has_key?(:custom_category_ids) then
            host.modify_categories2(
              host.sidfmd_categories.where(:is_system => true).pluck(:id).first,
              opts[:custom_category_ids],
              opts[:user_category_ids]
            )
            is_c_cats_update = true
          end
          if is_c_cats_update then
            change_category_ids_for_rules = SidfmdHost.get_change_category_ids(
              opts[:role_admin],
              opts[:user_category_ids],
              { host.hostname => opts[:custom_category_ids] }
            )
            raise Exception if change_category_ids_for_rules[:err]
            process_rules = SidfmdHost.process_related_rules(
              :role_admin => opts[:role_admin],
              :user_category_ids => opts[:user_category_ids],
              :action => 'category_change',
              :change_category_host_ids => change_category_ids_for_rules[:change_category_ids],
            )
            if !process_rules then
              raise Exception
            end
          end
        end
        # 基準日更新
        if is_host_admin_role && opts.has_key?(:refdate) then
          host.update_attribute :refdate, opts[:refdate]
        end
        # メモ更新
        if opts.has_key?(:memo) then
          host.update_attribute :memo, opts[:memo]
        end
        # アプリ更新
        if is_manual_products then
          if opts.has_key?(:host_apps) then
            host.sidfmd_host_apps = []
            opts[:host_apps].uniq!
            opts[:host_apps].each{|app| app.sidfmd_host_id = host.id }
            SidfmdHostApp.import(opts[:host_apps]) if opts[:host_apps].any?
          end
          if opts.has_key?(:host_lapps) then
            host.sidfmd_host_lapps = []
            opts[:host_lapps].uniq!
            opts[:host_lapps].each{|lapp| lapp.sidfmd_host_id = host.id }
            SidfmdHostLapp.import(opts[:host_lapps]) if opts[:host_lapps].any?
          end
        end

        # パッケージ更新
        if opts.has_key?(:package_os) && opts.has_key?(:packages) then
          update_host_packages(pkgs: opts[:packages], host: host, osname: opts[:package_os])
        end

        # ホストの検出アプリの有効/無効更新
        if opts.has_key?(:enable_apps) then
          opts[:enable_apps].each do |app|
            data = create_app_data.(host.id, app)
            SidfmdHostDisabledApp.where(data).destroy_all
          end
        end
        if opts.has_key?(:disable_apps) then
          disable_apps = opts[:disable_apps].map{|app|
            data = create_app_data.(host.id, app)
            SidfmdHostDisabledApp.new(data)
          }
          SidfmdHostDisabledApp.import(disable_apps) if disable_apps.any?
        end
      end

    rescue Exception => e
      log_exception(e)
      return create_process_result 'ホストの更新に失敗しました。'
    end
    if host then
      update_necessary = [
        :hostname,
        :display_name,
        :system_category_id,
        :custom_category_ids,
      ]
      .find{|item| opts.has_key?(item) }
      if update_necessary then
        # Viewに反映
        begin
          html_ids = SidfmdHostView.relate_htmlids([host.id]).uniq
          SidfmdHostView.sync_data([ host.id ])
          SidfmdVulView.sync_data(html_ids) if html_ids.any?
        rescue Exception => e
          logger.info('failed to update data related to view')
          log_exception(e)
        end
      end
    end
    Rails.cache.clear
    return create_process_result
  end

  # ホストの削除状態の変更
  # host_idはライセンスサーバのホストのID
  def toggle_delete_host(
    host_id: nil,
    sidfmd_host_id:,
    role_admin:,
    sidfm_roles:,
    user_host_ids:,
    user_category_ids:,
    update_view:
  )
    sidfmd_host_id = sidfmd_host_id.to_i
    if !sidfm_roles.include?(:host_admin) then
      return create_process_result 'forbidden access.'
    elsif !role_admin && !user_host_ids.include?(sidfmd_host_id) then
      return create_process_result 'forbidden access.'
    end

    host = nil
    begin
      SidfmdHost.transaction do
        host = SidfmdHost.lock.find_by(:id => sidfmd_host_id)
        unless host then
          res = SidfmdHost.establish_consistency
          logger.info("SidfmdHost.establish_consistensy=#{res.inspect}")
          return create_process_result '削除対象のホストがありません。'
        end
        unless host_id then
          license_host_data = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
          if license_host_data && license_host_data.has_key?(:hosts) then
            license_host = license_host_data[:hosts].find{|l_host| l_host[:sidfmd_host] && l_host[:sidfmd_host].id == host.id }
            if license_host then
              host_id = license_host[:id]
            end
          end
        end
        unless host_id then
          return create_process_result 'ライセンスサーバとの通信に失敗しました。'
        end
        new_flg_delete = !host[:flg_delete]
        host.update_attribute :flg_delete, new_flg_delete
        if new_flg_delete then
          process_rules = SidfmdHost.process_related_rules(
            :role_admin => role_admin,
            :user_category_ids => user_category_ids,
            :action => 'delete',
            :delete_host_ids => [ sidfmd_host_id ],
          )
          if !process_rules then
            raise Exception
          end
          # データを削除する。Redmineのチケットは削除しない（VM上からは直接参照できない）。
          last_checked_at = host.last_checked_at
          host.last_checked_at = nil
          host.memo = nil
          host.is_agent = false
          host.sidfmd_probe_apikey_id = nil
          host.probe_key = nil
          host.save
          # カテゴリの削除
          SidfmdCategoriesHost.where(:sidfmd_host_id => sidfmd_host_id).delete_all
          # プロダクトの削除
          SidfmdHostApp.where(:sidfmd_host_id => sidfmd_host_id).delete_all
          SidfmdHostLapp.where(:sidfmd_host_id => sidfmd_host_id).delete_all
          SidfmdHostDisabledApp.where(:sidfmd_host_id => sidfmd_host_id).delete_all
          # キューの削除
          sidfmd_queue_ids = SidfmdQueue.where(:sidfmd_host_id => sidfmd_host_id).collect {|x| x.id }.uniq
          sidfmd_queue_ids.each do |sidfmd_queue_id|
            SidfmdQueueDist.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            SidfmdQueuePkg.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            SidfmdQueueApp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            SidfmdQueueLapp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            SidfmdQueueWinApp.where(:sidfmd_queue_win_id => SidfmdQueueWin.where(:sidfmd_queue_id => sidfmd_queue_id).select(:id)).delete_all
            SidfmdQueueWin.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            sidfmd_queue_vul_ids = SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).collect {|x| x.id }.uniq
            sidfmd_queue_vul_ids.each do |sidfmd_queue_vul_id|
              SidfmdQueueVulCve.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
              SidfmdQueueVulSri.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
              SidfmdQueueVulAffprod.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
              SidfmdQueueVulAffver.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
              SidfmdQueueVulAffpkg.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
            end
            SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
          end
          SidfmdQueue.where(:id => sidfmd_queue_ids).delete_all
          # チケット（脆弱性）の削除
          sidfmd_issue_ids = SidfmdIssue.where(:sidfmd_host_id => sidfmd_host_id).collect {|x| x.id }.uniq
          sidfmd_issue_ids.each do |sidfmd_issue_id|
            SidfmdIssueCve.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            SidfmdIssueSri.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            SidfmdIssueAffprod.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            SidfmdIssueAffver.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            SidfmdIssueAffpkg.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            SidfmdIssueCtag.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
          end
          SidfmdIssue.where(:id => sidfmd_issue_ids).delete_all
        end
        res = SidfmdHost.toggle_host_delete(host_id.to_i)
        if !res || !res[:success] then
          logger.info('failed to access license server.')
          raise Exception
        end
      end
    rescue Exception => e
      log_exception(e)
      return create_process_result 'ホストの削除に失敗しました。'
    end
    if update_view then
      # API等、すぐに更新が必要な場合にViewを更新
      if host then
        begin
          html_ids = SidfmdHostView.relate_htmlids([host.id]).uniq
          SidfmdHostView.sync_data([ host.id ])
          SidfmdVulView.sync_data(html_ids) if html_ids.any?
        rescue Exception => e
          logger.info('failed to update data related to view')
          log_exception(e)
        end
      end
    end
    Rails.cache.clear
    return create_process_result
  end

  # ホストのパッケージ更新
  # 呼び出し元でエラーハンドリング等する
  def update_host_packages(pkgs:, host:, osname:)

    validation = validate_packages(pkgs: pkgs, osname: osname)
    unless validation[:success] then
      return create_process_result validation[:message]
    end
    pkgs.uniq!

    host.update(:is_agent => true) unless host.is_agent
    queue = host.sidfmd_queues.order(:created_at => :desc).first
    data = { :packages => pkgs }
    os_data = nil
    if queue then
      queue_os = queue.sidfmd_queue_dist
      if queue_os then
        #最新のSidfmdQueueDistが選択されたものと同じで、且つ更新可能なパッケージのOSなら、
        #SidfmdQueueDistに登録するOSは最新のSidfmdQueueDistにする
        queue_os_name = queue_os.name
        if queue_os_name.match(/RedHat Enterprise Linux/) then #DBにRed Hatの「Red」と「Hat」の間に空白が無いものがあるので、念のため補正
          queue_os_name = 'Red Hat Enterprise Linux'
        end
        if osname.match(queue_os_name) then
          selected_version = if /Ubuntu/ =~ osname then
                               osname.slice(/\d+\.\d+/)
                             else
                               osname.slice(/\d+/)
                             end
          if queue_os.version.index(selected_version) then
            begin
              os_data = SidfmdQueueDist.get_updatable_pkg_os_data("#{queue_os.name} #{queue_os.version}")
            rescue Exception => e
              #更新不可のバージョンだった場合
              os_data = nil
            end
          end
        end
      end
      queue_apps = queue.sidfmd_queue_apps
      data[:applications] = (queue_apps.any? ? queue_apps.map{|app| [ app.name, app.version, app.path ]} : nil)
    end
    os_data = SidfmdQueueDist.get_updatable_pkg_os_data(osname) unless os_data
    data[:os] = os_data
    host.create_queue(data)
  end

  # for search product name
  def searchProduct(key, product_names, version_names, file_name="dummy file", line_num="dummy line num", product_type)
    ret = {}
    product_names_ik = product_names.transform_keys{ |k| k.downcase }
    version_names_ik = version_names.transform_keys{ |k| k.downcase }

    unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids

    if version_names[key] then
      ret = { success: true, value: version_names[key], type: "version" }
    else
      if version_names_ik[key.downcase] then
        ret = { success: true, value: version_names_ik[key.downcase], type: "version" }
      else
        # search with key and version. this maybe most difficult part
        temp = key.downcase.split(" ")
        version_part = temp[-1]
        if temp.count > 1 && version_part =~ /^[\d\.x]+$/ then
          count=version_part.count(".")
          if count > 0 then
            for i in 0..count-1 do
              mod_ver =version_part.gsub(/([\dx]+.){#{i}}[\dx]+$/, "x")
              if mod_ver == version_part then
                next
              end
              temp[-1] = mod_ver
              re = searchProduct(temp.join(" "), product_names, version_names, file_name, line_num, product_type)
              if re.empty? then
                next
              elsif re[:success] == false then
                next
              else
                ret = re
                break
              end
            end
          end
          if ret.empty? then
            ret = { success: false, err: "#{file_name}の#{line_num}行目: #{product_type}の「データ」列の値に、登録されていないプロダクト名が入力されています。" }
          end
        else
          # check product name
          if product_names[key] then
            if product_type == "manual_products" then

              # check if product is registable or not
              pid = SidfmdProduct.find_by(:name => key).pid
              if unregistrable_product_ids.include?(pid) then
                ret = { success: false, err: "#{file_name}の#{line_num}行目: #{product_type}の「データ」列の値に、 仕様によりバージョンを指定する場合にのみしか登録することができないプロダクト名が入力されています。" }
              else
                ret = { success: true, value: product_names[key], type: "product"}
              end
            else
              # register if product type is local_products
              ret = { success: true, value: product_names[key], type: "product"}
            end
          else
            # re-search product name with case insensitive
            if product_names_ik[key.downcase] then
              org_key = product_names.key(product_names_ik[key.downcase])
              if product_type == "manual_products" then
                # check if unregistable product or not
                pid = SidfmdProduct.find_by(:name => org_key).pid
                if unregistrable_product_ids.include?(pid) then
                  Rails.logger.info("unregistable!!!!!")
                  ret = { success: false, err: "#{file_name}の#{line_num}行目: #{product_type}の「データ」列の値に、 仕様によりバージョンを指定する場合にのみしか登録することができないプロダクト名が入力されています。" }
                else
                  ret = { success: true, value: product_names[org_key], type: "product"}
                end
              else
                ret = { success: true, value: product_names[org_key], type: "product"}
              end
            else
              # not found case
              Rails.logger.info("Searched product name with ignore case, but nothing was found")
              ret = { success: false, err: "#{file_name}の#{line_num}行目: #{product_type}の「データ」列の値に、登録されていないプロダクト名が入力されています。" }
            end
          end
        end
      end
    end
    ret
  end

  def removeDuplication(set_name, value)
    logger.info("check duplication")
    if !set_name.include?(value) then
      set_name.add(value)
    end
  end

end

